package eel6788.agents.directeddiffusion;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import yaes.framework.agent.ACLMessage;
import yaes.framework.world.sensornetwork.Perception;
import yaes.framework.world.sensornetwork.SensingHistory;
import yaes.framework.world.sensornetwork.SensorNetworkWorld;
import yaes.ui.text.TextUi;
import eel6788.MessageConstants;
import eel6788.agents.SimpleSensorAgent;

/**
 * 
 * This agent implements a very simple version of the directed diffusion
 * paradigm. It knows its own position and the position of the sink. If a
 * message is coming from a node which is closer to the sink than him, it does
 * not retransmit it. Otherwise it transmits it.
 * 
 * If it observes a node, it immediately transmits its position.
 * 
 * 
 * @author Lotzi Boloni
 * 
 */
public class DirectedDiffusionAgent extends SimpleSensorAgent
        implements
            MessageConstants {

    private final HashMap<String, Interest> interests = new HashMap<String, Interest>();

    public DirectedDiffusionAgent(String name, SensorNetworkWorld sensingManager) {
        super(name, sensingManager);
    }

    /**
     * Processes the perceptions:
     * <ul>
     * <li> actuator presence
     * <li> received message of interest
     * <li> received message
     * <li> data
     * </ul>
     * 
     * Also create empty messages if there is nothing to send.
     * 
     */
    @Override
    public void action() {
        super.action();
        // read out his stuff from the sensing manager
        final SensingHistory sensingHistory = getSensorWorld()
                .getSensingHistory(getNode());
        final List<Perception> perceptions = sensingHistory.getMailbox();
        boolean hadObservation = false;
        for (final Perception p : perceptions) {
            switch (p.getType()) {
                case ActuatorPresence :
                    processObservation(p);
                    hadObservation = true;
                    break;
                case ReceivedMessage :
                    final ACLMessage m = p.getMessage();
                    if (m.getValue(MessageConstants.FIELD_CONTENT)
                            .equals(MessageConstants.MESSAGE_INTEREST)) {
                        processInterestMessage(m);
                        continue;
                    }
                    if (m.getValue(MessageConstants.FIELD_CONTENT)
                            .equals(MessageConstants.MESSAGE_DATA)) {
                        processDataMessage(m);
                        continue;
                    }
                    break;
                case Overhearing :
                    continue;
            }
        }
        // if there was no observation, maybe we need to send some messages
        // anyhow
        if (hadObservation == false) {
            processNoObservation();
        }
        // finally, process the time
        final ArrayList<Interest> toDelete = new ArrayList<Interest>();
        for (final Interest interest : interests.values()) {
            interest.updateInterestBasedOnTime(getSensorWorld().getTime());
            if (interest.getGradientCollection().isEmpty()) {
                toDelete.add(interest);
            }
        }
        // remove the expired interests
        for (final Interest interest : toDelete) {
            interests.remove(interest.getInterestType());
        }
    }

    /**
     * 
     * An observation was performed.
     * 
     * <ul>
     * <li> Check for an interest in the type of observation.
     * <li> Check for gradients on the observation.
     * <li> For each gradient if it is time to send, send a peer to peer
     * message.
     * </ul>
     * 
     * @param p
     */
    private void processObservation(Perception p) {
        for (final Interest interest : interests.values()) {
            if (interest.isLocationOfInterest(node.getLocation())) {
                for (final Gradient gradient : interest.getGradientCollection()) {
                    final int lastMessageSent = gradient.getLastMessageSent();
                    final int currentTime = getSensorWorld().getTime();
                    if (currentTime - lastMessageSent < gradient.getInterval()) {
                        continue;
                    }
                    gradient.setLastMessageSent(currentTime);
                    ACLMessage message = createMessage(interest, gradient, p);
                    transmit(message);
                }
            }
        }
    }

    /**
     * 
     * No observation was performed.
     * 
     * <ul>
     * <li> Check for an interest in the type of observation.
     * <li> Check for gradients on the observation.
     * <li> For each gradient if it is time to send, send a peer to peer
     * message.
     * </ul>
     * 
     * @param p
     */
    private void processNoObservation() {
        for (final Interest interest : interests.values()) {
            Perception p = null;
            if (interest.isLocationOfInterest(node.getLocation())) {
                for (final Gradient gradient : interest.getGradientCollection()) {
                    final int lastEmptyMessageSent = gradient
                            .getLastEmptyMessageSent();
                    final int currentTime = getSensorWorld().getTime();
                    if (currentTime - lastEmptyMessageSent < gradient
                            .getInterval()) {
                        continue;
                    }
                    // it is time to send something:
                    if (p == null) {
                        p = new Perception(
                                Perception.PerceptionType.NoPerception,
                                currentTime);
                    }
                    gradient.setLastEmptyMessageSent(currentTime);
                    ACLMessage message = createMessage(interest, gradient, p);
                    transmit(message);
                }
            }
        }
    }

    /**
     * Creates a message for a gradient based on an observation. If the
     * perception is NoPerception, then it will send a message with the
     * perception put to 0.
     * 
     * @return
     */
    private ACLMessage createMessage(Interest interest, Gradient gradient,
            Perception p) {
        final ACLMessage message = new ACLMessage(getNode().getName(),
                ACLMessage.Performative.INFORM);
        message.setDestination(gradient.getSensorName());
        message.setValue(MessageConstants.FIELD_CONTENT,
                         MessageConstants.MESSAGE_DATA);
        message.setValue(MessageConstants.FIELD_TYPE, interest
                .getInterestType());
        message.setValue(MessageConstants.FIELD_LOCATION, getNode()
                .getLocation());
        message.setValue(MessageConstants.FIELD_PERCEPTION_ID, p.getId());
        if (p.getType() != Perception.PerceptionType.NoPerception) {
            message.setValue(MessageConstants.FIELD_INTENSITY, 1);
        } else {
            message.setValue(MessageConstants.FIELD_INTENSITY, 0);
        }
        return message;
    }

    /**
     * An data message was received.
     * 
     * <ul>
     * <li> check for an interest in the type of observation
     * <li> check for gradients.
     * <li> for each gradient, if it is time to send, send a peer-to-peer
     * message
     * </ul>
     * 
     * @param m
     */
    private void processDataMessage(ACLMessage m) {
        final String interestType = (String) m
                .getValue(MessageConstants.FIELD_TYPE);
        final Interest interest = interests.get(interestType);
        if (interest == null) { // no interest, do not do anything
            return;
        }
        final int perceptionId = (Integer) m
                .getValue(MessageConstants.FIELD_PERCEPTION_ID);
        String sender = m.getSender();
        if (interest.processPerceptionId(perceptionId, sender)) {
            // the perception was previously processed.
            return;
        }
        for (final Gradient gradient : interest.getGradientCollection()) {
            final ACLMessage retransmit = new ACLMessage(getNode().getName(),
                    ACLMessage.Performative.INFORM);
            retransmit.setDestination(gradient.getSensorName());
            retransmit.setValue(MessageConstants.FIELD_PERCEPTION_ID,
                                perceptionId);
            retransmit.setValue(MessageConstants.FIELD_CONTENT,
                                MessageConstants.MESSAGE_DATA);
            retransmit.setValue(MessageConstants.FIELD_TYPE, interest
                    .getInterestType());
            retransmit.setValue(MessageConstants.FIELD_LOCATION, m
                    .getValue(MessageConstants.FIELD_LOCATION));
            retransmit.setValue(MessageConstants.FIELD_INTENSITY, m
                    .getValue(MessageConstants.FIELD_INTENSITY));
            transmit(retransmit);
        }
    }

    /**
     * An interest was received.
     * <ul>
     * <li> check whether there is an interest for the type. If not, create.
     * <li> check whether there is a gradient for the sender. If not, create.
     * <li> update the gradient.
     * <li> forward the interest if any update was made
     * </ul>
     * 
     * @param m
     */
    private void processInterestMessage(ACLMessage message) {
        boolean forward = false;
        final String interestType = (String) message
                .getValue(MessageConstants.FIELD_TYPE);
        final String interestReceivedFromSensor = message.getSender();
        final Rectangle2D.Double interestRectangle = (Rectangle2D.Double) message
                .getValue(MessageConstants.FIELD_RECTANGLE);
        final int interestDuration = (Integer) message
                .getValue(MessageConstants.FIELD_DURATION);
        final int interestInterval = (Integer) message
                .getValue(MessageConstants.FIELD_INTERVAL);
        // update interests
        Interest interest = interests.get(interestType);
        boolean reinforcement = false;
        if (interest == null) { // create the interest
            interest = new Interest(interestType, interestReceivedFromSensor,
                    interestInterval, interestDuration, interestRectangle);
            interests.put(interestType, interest);
            forward = true;
        } else { // look up the gradient
            forward = interest
                    .updateInterestOnNewMessage(interestReceivedFromSensor,
                                                interestInterval,
                                                interestDuration);
            reinforcement = true;
        }
        if (!forward) {
            return;
        }
        if (reinforcement == false) {
            ACLMessage forwardMessage = createForwardMessage("*", interestType,
                                                             interestRectangle,
                                                             interestDuration,
                                                             interestInterval);
            transmit(forwardMessage);
        } else {
            doReinforcement(interest, interestDuration, interestInterval);
        }
    }

    /**
     * Does the reinforcement. For each interest, choose the node to reinforce.
     * These are the nodes which were the first to report.
     */
    public void doReinforcement(Interest interest, int interestDuration,
            int interestInterval) {
        Set<String> toReinforce = new HashSet<String>();
        Map<Integer, List<String>> processedPerceptions = interest
                .getProcessedPerceptions();
        for (List<String> reporters : processedPerceptions.values()) {
            toReinforce.add(reporters.get(0));
        }
        if (toReinforce.isEmpty()) {
            TextUi.println("Trouble: " + interest + " did not get any report");
            return;
        }
        for (String dest : toReinforce) {
            ACLMessage message = createForwardMessage(dest, interest
                    .getInterestType(), interest.getRectangle(),
                                                      interestDuration,
                                                      interestInterval);
            transmit(message);
        }
    }

    private ACLMessage createForwardMessage(String destination,
            String interestType, Rectangle2D.Double interestRectangle,
            int interestDuration, int interestInterval) {
        final ACLMessage forwardMessage = new ACLMessage(getName(),
                ACLMessage.Performative.REQUEST_WHENEVER);
        forwardMessage.setDestination(destination);
        forwardMessage.setValue(MessageConstants.FIELD_CONTENT,
                                MessageConstants.MESSAGE_INTEREST);
        forwardMessage.setValue(MessageConstants.FIELD_TYPE, interestType);
        forwardMessage.setValue(MessageConstants.FIELD_RECTANGLE,
                                interestRectangle);
        forwardMessage.setValue(MessageConstants.FIELD_DURATION,
                                interestDuration);
        forwardMessage.setValue(MessageConstants.FIELD_INTERVAL,
                                interestInterval);
        return forwardMessage;
    }

    /**
     * Process the passing of time
     * 
     * @param time
     */
    private void processTimePassing(int time) {
    }

    /**
     * Prints out an agent for debugging purposes
     */
    @Override
    public String toString() {
        final StringBuffer buffer = new StringBuffer("DirectedDiffusionAgent: "
                + getName());
        if (!interests.isEmpty()) {
            for (final Interest interest : interests.values()) {
                buffer.append("\n" + interest);
            }
        } else {
            buffer.append("\n No interests.");
        }
        buffer.append("\n");
        return buffer.toString();
    }
}
