package eel6788.agents.directeddiffusion;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Formatter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import yaes.framework.agent.ACLMessage;
import yaes.framework.world.sensornetwork.AbstractSensorAgent;
import yaes.framework.world.sensornetwork.Perception;
import yaes.framework.world.sensornetwork.SensingHistory;
import yaes.framework.world.sensornetwork.SensorNetworkWorld;
import yaes.ui.text.TextUi;
import eel6788.MessageConstants;

public class DirectedDiffusionSinkAgent extends AbstractSensorAgent implements
		MessageConstants {
	private static final int INITIAL_INTERVAL = 5;
	private static final int INITIAL_DURATION = 30;
	private static final int EVALUATION_TIME = 20;
	private static final int FINAL_INTERVAL = 5;
	private static final int FINAL_DURATION = 3000;
	private int reinforcementStarted = -1;
	private int receivedDataCount = 0;
	private final HashMap<String, SinkInterest> interests =
			new HashMap<String, SinkInterest>();
	private List<Rectangle2D.Double> interestRectangles =
			new ArrayList<Rectangle2D.Double>();

	public DirectedDiffusionSinkAgent(String name,
			SensorNetworkWorld sensingManager) {
		super(name, sensingManager);
		// cover the area 200, 200 -->  600, 600
		int xstep = 50;
		int ystep = 50;
		for(int x = 200; x != 600; x = x + xstep) {
			for(int y = 200; y != 600; y = y + ystep) {
				interestRectangles.add(new Rectangle2D.Double(x, y, xstep, ystep));				
			}
		}
		// create the interests
		for (Rectangle2D.Double rect : interestRectangles) {
			String label = getInterestLabel(rect);
			SinkInterest interest = new SinkInterest(label, rect);
			interests.put(interest.getInterestType(), interest);
		}
	}

	/**
	 * 
	 * The sink attends to all the interests and takes appropriate actions.
	 * 
	 * The sink did not yet expressed any interest. It starts by sending out a
	 * message and expressing interest for every interval.
	 */
	public void action() {
		actionListenToMessages();
		for (SinkInterest si : interests.values()) {
			attendToInterest(si);
		}
	}

	private void attendToInterest(SinkInterest si) {
		// the interest was not yet expressed, express it
		if (si.getDuration() == 0) {
			si.setDuration(getSensorWorld().getTime() + INITIAL_DURATION);
			si.setInterval(INITIAL_INTERVAL);
			ACLMessage message =
					createInterestMessage("*", si.getRectangle(), si
							.getDuration(), si.getInterval());
			transmit(message);
		}
		// the interest is in the initial reinforcement state
		// and half the evaluation time has passed
		if ((si.getInterval() == INITIAL_INTERVAL) && 
				(si.getDuration() - getSensorWorld().getTime() < (EVALUATION_TIME / 2))) {
			doReinforcement(si);
		}

	}

	/**
	 * Creates an interest message for a rectangle, duration, interval
	 * 
	 * @param rect
	 * @param duration
	 * @param interval
	 * @return
	 */
	private ACLMessage createInterestMessage(String destination,
			Rectangle2D.Double rect, int duration, int interval) {
		final ACLMessage message =
				new ACLMessage(getName(),
						ACLMessage.Performative.REQUEST_WHENEVER);
		message.setDestination(destination);
		message.setValue(MessageConstants.FIELD_CONTENT,
				MessageConstants.MESSAGE_INTEREST);
		message.setValue(MessageConstants.FIELD_TYPE, getInterestLabel(rect));
		message.setValue(MessageConstants.FIELD_RECTANGLE, rect);
		message.setValue(MessageConstants.FIELD_DURATION, duration);
		message.setValue(MessageConstants.FIELD_INTERVAL, interval);
		return message;
	}

	/**
	 * Returns a human readable label for a specific interest. It will be based
	 * on the rectangle.
	 * 
	 * @param rectangle
	 * @return
	 */
	private String getInterestLabel(Rectangle2D.Double rectangle) {
		StringBuilder sb = new StringBuilder();
		Formatter formatter = new Formatter(sb);
		formatter.format("%s_%.0f_%.0f_%.0f_%.0f",
				MessageConstants.INTEREST_TYPE_INTRUDER, rectangle.getMinX(),
				rectangle.getMinY(), rectangle.getMaxX(), rectangle.getMaxY());
		return sb.toString();
	}


	/**
	 * Does the reinforcement. For each interest, choose the node to reinforce.
	 * These are the nodes which were the first to report.
	 */
	public void doReinforcement(SinkInterest interest) {
		Set<String> toReinforce = new HashSet<String>();
		Map<Integer, List<String>> processedPerceptions =
				interest.getProcessedPerceptions();
		for (List<String> reporters : processedPerceptions.values()) {
			toReinforce.add(reporters.get(0));
		}
		if (toReinforce.isEmpty()) {
			TextUi.println("Trouble: " + interest + " did not get any report");
			return;
		}
		for (String dest : toReinforce) {
			ACLMessage message =
					createInterestMessage(dest, interest.getRectangle(),
							FINAL_DURATION, FINAL_INTERVAL);
			transmit(message);
		}
	}

	/**
	 * Listen to incoming messages and maintain the values in the interests.
	 * 
	 */
	private void actionListenToMessages() {
		final SensingHistory sensingHistory =
				getSensorWorld().getSensingHistory(getNode());
		final List<Perception> perceptions = sensingHistory.getMailbox();
		for (final Perception p : perceptions) {
			if (p.getType() == Perception.PerceptionType.ActuatorPresence) {
				// ignore direct perceptions
				continue;
			}
			if (p.getType() == Perception.PerceptionType.ReceivedMessage) {
				final ACLMessage m = p.getMessage();
				if (m.getValue(MessageConstants.FIELD_CONTENT).equals(
						MessageConstants.MESSAGE_INTEREST)) {
					// ignore interest messages
					continue;
				}
				if (m.getValue(MessageConstants.FIELD_CONTENT).equals(
						MessageConstants.MESSAGE_DATA)) {
					processReceivedData(m);
				}
			}
		}
	}

	/**
	 * Processes the received data.
	 * 
	 * @param m
	 */
	private void processReceivedData(ACLMessage m) {
		final String interestType =
				(String) m.getValue(MessageConstants.FIELD_TYPE);
		final SinkInterest interest = interests.get(interestType);
		if (interest == null) { // no interest, do not do anything
			return;
		}
		final int perceptionId =
				(Integer) m.getValue(MessageConstants.FIELD_PERCEPTION_ID);
		String sender = m.getSender();
		interest.processPerceptionId(perceptionId, sender);
		receivedDataCount++;
	}

	/**
	 * Returns a description of the state of the agent
	 */
	@Override
	public String toString() {
		final StringBuffer buffer =
				new StringBuffer("Directed Diffusion Sink ");
		buffer.append("Interests:");
		for (SinkInterest interest : interests.values()) {
			buffer.append("\n"
					+ TextUi.shiftTextWithSpaces(3, interest.toString()));
		}
		return buffer.toString();
	}

	public HashMap<String, SinkInterest> getInterests() {
		return interests;
	}
}
